from fiscalsim_us.model_api import *
from policyengine_core.parameters import homogenize_parameter_structures
from policyengine_core.simulations import Simulation
from fiscalsim_us.variables.household.demographic.geographic.state_name import (
    StateName,
)

label = "Geography"


class state_name(Variable):
    value_type = Enum
    possible_values = StateName
    default_value = StateName.CA
    entity = Household
    label = "State"
    definition_period = YEAR

    def formula(household, period, parameters):
        fips = household("state_fips", period)
        return (
            pd.Series(fips)
            .map(
                {
                    1: StateName.AL,
                    2: StateName.AK,
                    4: StateName.AZ,
                    5: StateName.AR,
                    6: StateName.CA,
                    8: StateName.CO,
                    9: StateName.CT,
                    10: StateName.DE,
                    11: StateName.DC,
                    12: StateName.FL,
                    13: StateName.GA,
                    15: StateName.HI,
                    16: StateName.ID,
                    17: StateName.IL,
                    18: StateName.IN,
                    19: StateName.IA,
                    20: StateName.KS,
                    21: StateName.KY,
                    22: StateName.LA,
                    23: StateName.ME,
                    24: StateName.MD,
                    25: StateName.MA,
                    26: StateName.MI,
                    27: StateName.MN,
                    28: StateName.MS,
                    29: StateName.MO,
                    30: StateName.MT,
                    31: StateName.NE,
                    32: StateName.NV,
                    33: StateName.NH,
                    34: StateName.NJ,
                    35: StateName.NM,
                    36: StateName.NY,
                    37: StateName.NC,
                    38: StateName.ND,
                    39: StateName.OH,
                    40: StateName.OK,
                    41: StateName.OR,
                    42: StateName.PA,
                    44: StateName.RI,
                    45: StateName.SC,
                    46: StateName.SD,
                    47: StateName.TN,
                    48: StateName.TX,
                    49: StateName.UT,
                    50: StateName.VT,
                    51: StateName.VA,
                    53: StateName.WA,
                    54: StateName.WV,
                    55: StateName.WI,
                    56: StateName.WY,
                    72: StateName.PR,
                }
            )
            .values
        )


class medicaid_rating_area(Variable):
    value_type = int
    entity = Household
    label = "Medicaid rating area"
    definition_period = YEAR
    hidden_input = True

    def formula(household, period, parameters):
        simulation: Simulation = household.simulation
        if (
            simulation.get_holder("reported_slspc").get_array(period)
            is not None
        ):
            # If the user has provided a value for the second-lowest silver plan
            # cost, skip.
            return 0
        parameter_tree = household.simulation.tax_benefit_system.parameters
        if not hasattr(parameter_tree.gov.hhs.medicaid, "geography"):
            medicaid_parameters = ParameterNode(
                directory_path=REPO
                / "data"
                / "parameters"
                / "gov"
                / "hhs"
                / "medicaid"
                / "geography"
            )
            medicaid_parameters = homogenize_parameter_structures(
                medicaid_parameters,
                household.simulation.tax_benefit_system.variables,
            )
            parameter_tree.gov.hhs.medicaid.add_child(
                "geography", medicaid_parameters
            )
        mra = parameter_tree.gov.hhs.medicaid.geography.medicaid_rating_area(
            period
        )
        three_digit_zip_code = household("three_digit_zip_code", period)
        county = household("county_str", period)
        locations = np.array(list(mra._children))
        county_in_locations = np.isin(county, locations)
        location = where(
            county_in_locations,
            county,
            three_digit_zip_code,
        )
        valid_location = np.isin(location, locations)
        rating_areas = np.ones_like(
            location
        )  # ~2.8% of locations don't match with a a scraped MRA. For these, we assign to the State's first MRA.
        if valid_location.sum() > 0:
            rating_areas[valid_location] = mra[location[valid_location]]
        return rating_areas


class reported_slspc(Variable):
    value_type = float
    entity = TaxUnit
    label = "reported second lowest silver plan cost"
    unit = USD
    definition_period = YEAR
    hidden_input = True


class county_fips(Variable):
    value_type = int
    entity = Household
    definition_period = YEAR
    documentation = "County FIPS code"


class state_fips(Variable):
    value_type = int
    entity = Household
    definition_period = YEAR
    documentation = "State FIPS code"
    default_value = 6



from fiscalsim_us.model_api import *


class mn_itemized_deductions(Variable):
    """
    Line 4 of Minnesota 2022 Individual Income Tax return from M1.
    """

    value_type = float
    entity = TaxUnit
    label = "MN itemized deductions"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MN
    reference = (
        "https://www.revenue.state.mn.us/sites/default/files/2021-12/m1_21_0.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-01/m1_inst_21.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2022-12/m1_22.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-03/m1_inst_22.pdf"
    )

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.mn.tax.income.deductions

        fed_agi = tax_unit("adjusted_gross_income", period)
        filing_status = tax_unit("filing_status", period)

        # deductions that are limited to 20% of the deduction over the
        # income threshold
        limited_deductions = [
            "mn_taxes_paid_deducion",
            "charitable_deduction",
            "mn_unreimbursed_employee_deduction",
            "mn_gambling_loss_deduction",
            "mn_disabled_impairment_work_deduction",
            "mn_home_mortgage_interest_deduction",
            "mn_other_itemized_deductions",
        ]

        # deductions that aren't limited over the income threshold
        # (they don't get phased out)
        non_limited_deductions = [
            "mn_medical_dental_deduction",
            "interest_deduction",
            "mn_casualty_theft_deduction",
        ]

        limited_amount = add(tax_unit, period, limited_deductions)
        all_deductions = limited_amount + add(
            tax_unit, period, non_limited_deductions
        )

        return where(
            fed_agi <= p.itemized_threshold[filing_status],
            all_deductions,
            all_deductions
            - min_(
                limited_amount * p.itemized_mult,
                (fed_agi - p.itemized_threshold[filing_status])
                * p.itemized_income_mult,
            ),
        )



from fiscalsim_us.model_api import *


class self_employed_health_insurance_premiums(Variable):
    value_type = float
    entity = Person
    label = "Self-employed health insurance premiums"
    unit = USD
    documentation = "Health insurance premiums for plans covering individuals who are not covered by any employer-sponsored health insurance."
    definition_period = YEAR

    def formula(person, period, parameters):
        is_self_employed = person("is_self_employed", period)
        health_insurance_premiums = person("health_insurance_premiums", period)
        return is_self_employed * health_insurance_premiums



from fiscalsim_us.model_api import *


class is_tanf_continuous_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    definition_period = YEAR
    label = "Continued Economic Eligibility for TANF"
    documentation = "Whether the familiy meets the economic requirements for the Temporary Assistance for Needy Families program after being approved."

    def formula(spm_unit, period, parameters):
        earned_income = spm_unit("tanf_gross_earned_income", period)
        unearned_income = spm_unit("tanf_gross_unearned_income", period)
        net_income = earned_income + unearned_income
        payment_level = spm_unit("tanf_max_amount", period)
        return net_income <= payment_level



from fiscalsim_us.model_api import *


class de_additional_standard_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Delaware additional standard deduction"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://delcode.delaware.gov/title30/c011/sc02/index.html#1108"
    )
    defined_for = StateCode.DE

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.de.tax.income.deductions.standard.additional

        age_head = tax_unit("age_head", period)
        aged_head_eligible = (age_head >= p.age_threshold).astype(int)

        age_spouse = tax_unit("age_spouse", period)
        aged_spouse_eligible = (age_spouse >= p.age_threshold).astype(int)

        blind_head = tax_unit("blind_head", period)
        blind_head_eligible = (blind_head).astype(int)

        blind_spouse = tax_unit("blind_spouse", period)
        blind_spouse_eligible = (blind_spouse).astype(int)

        return (
            aged_head_eligible
            + aged_spouse_eligible
            + blind_head_eligible
            + blind_spouse_eligible
        ) * p.amount



from fiscalsim_us.model_api import *


class cliff_evaluated(Variable):
    value_type = bool
    entity = Person
    label = "cliff evaluated"
    unit = USD
    documentation = "Whether this person's cliff has been simulated. If not, the cliff gap is assumed to be zero."
    definition_period = YEAR

    def formula(person, period, parameters):
        adult_index_values = person("adult_index", period)
        cliff_adult_count = parameters(period).simulation.cliff_adults
        is_adult = person("is_adult", period)
        return is_adult & (adult_index_values <= cliff_adult_count)


class cliff_gap(Variable):
    value_type = float
    entity = Person
    label = "cliff gap"
    unit = USD
    documentation = "Amount of income lost if this person's employment income increased by delta amount."
    definition_period = YEAR

    def formula(person, period, parameters):
        netinc_base = person.household("household_net_income", period)
        delta = parameters(period).simulation.cliff_delta
        adult_count = parameters(period).simulation.cliff_adults
        sim = person.simulation
        increase = np.zeros(person.count, dtype=np.float32)
        adult_indexes = person("adult_index", period)
        for adult_index in range(1, 1 + adult_count):
            alt_sim = sim.get_branch(f"cliff_for_adult_{adult_index}")
            for variable in sim.tax_benefit_system.variables:
                if variable not in sim.input_variables:
                    alt_sim.delete_arrays(variable)
            mask = adult_index == adult_indexes
            alt_sim.set_input(
                "employment_income",
                period,
                person("employment_income", period) + mask * delta,
            )
            alt_person = alt_sim.person
            netinc_alt = alt_person.household("household_net_income", period)
            increase[mask] += netinc_alt[mask] - netinc_base[mask]
        return where(increase < 0, -increase, 0)


class is_on_cliff(Variable):
    value_type = bool
    entity = Person
    label = "is on a tax-benefit cliff"
    documentation = "Whether this person would be worse off if their employment income were higher."
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("cliff_gap", period) > 0



from fiscalsim_us.model_api import *


class va_exemptions(Variable):
    value_type = float
    entity = TaxUnit
    label = "VA exemptions"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.VA

    def formula(tax_unit, period, parameters):
        va_indiv_exempt_multiplier = parameters(
            period
        ).gov.states.va.tax.income.exemptions.indiv_exemption_multiplier
        va_indiv_p65_exempt_multiplier = parameters(
            period
        ).gov.states.va.tax.income.exemptions.indiv_exemption_over_65_multiplier

        dependents = tax_unit("tax_unit_dependents", period)
        filing_status = tax_unit("filing_status", period)
        filing_statuses = filing_status.possible_values

        spouse_if_filing_jointly = where(
            (filing_status == filing_statuses.SINGLE)
            | (filing_status == filing_statuses.HEAD_OF_HOUSEHOLD)
            | (filing_status == filing_statuses.SURVIVING_SPOUSE)
            | (filing_status == filing_statuses.SEPARATE),
            0,
            where(filing_status == filing_statuses.JOINT, 1, 0),
        )

        you_p65 = (tax_unit("age_head", period) >= 65).astype(int)
        spouse_p65 = (tax_unit("age_spouse", period) >= 65).astype(int)
        you_blind = tax_unit("blind_head", period).astype(int)
        spouse_blind = tax_unit("blind_spouse", period).astype(int)

        total_section_A = (
            1 + spouse_if_filing_jointly + dependents
        ) * va_indiv_exempt_multiplier
        total_section_B = (
            you_p65 + spouse_p65 + you_blind + spouse_blind
        ) * va_indiv_p65_exempt_multiplier

        return total_section_A + total_section_B



from fiscalsim_us.model_api import *


class mt_investment_interest_deduction(Variable):
    """
    Line 10 on itemized deductions schedule
    """

    value_type = float
    entity = TaxUnit
    label = "Montana investment interest deduction"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT

    def formula(tax_unit, period, parameters):
        contributions = tax_unit("investment_income_form_4952")

        return contributions



from fiscalsim_us.model_api import *


# TODO: apparently test is required for this, even though this is directly linked to il_personal_exemption_eligibility_status?
class il_personal_exemption(Variable):
    value_type = float
    entity = TaxUnit
    label = "IL personal exemption amount"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        il_is_personal_exemption_eligible = tax_unit(
            "il_personal_exemption_eligibility_status", period
        )

        eligibility_status = il_is_personal_exemption_eligible.possible_values

        personal_exemption_amount = parameters(
            period
        ).gov.states.il.tax.income.exemption.personal

        return personal_exemption_amount * select(
            [
                il_is_personal_exemption_eligible
                == eligibility_status.BOTH_ELIGIBLE,
                il_is_personal_exemption_eligible
                == eligibility_status.PARTIALLY_ELIGIBLE,
            ],
            [2, 1],
            0,
        )



from fiscalsim_us.model_api import *


class ami(Variable):
    value_type = float
    entity = SPMUnit
    label = "Area median income"
    documentation = "Area median income"
    definition_period = YEAR



from fiscalsim_us.model_api import *


class dc_tanf_countable_income(Variable):
    value_type = float
    entity = SPMUnit
    label = "DC TANF countable income"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.DC

    def formula(spm_unit, period, parameters):
        gross_earnings = spm_unit("dc_tanf_gross_earned_income", period)
        gross_unearned_income = spm_unit(
            "dc_tanf_countable_gross_unearned_income", period
        )
        p = parameters(period).gov.states.dc.dhs.tanf.income.deductions.earned
        annual_flat_exclusion = p.flat * MONTHS_IN_YEAR
        earnings_after_deduction = max_(
            gross_earnings - annual_flat_exclusion, 0
        ) * (1 - p.percentage)
        return earnings_after_deduction + gross_unearned_income



from fiscalsim_us.model_api import *


class tax_unit_medicaid_income_level(Variable):
    value_type = float
    entity = TaxUnit
    label = "Medicaid income level"
    unit = "/1"
    documentation = (
        "Income for Medicaid as a percentage of the federal poverty line."
    )
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        income = tax_unit("medicaid_income", period)
        fpg = tax_unit("tax_unit_fpg", period)
        return income / fpg



from fiscalsim_us.model_api import *


class ri_property_tax_credit_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Rhode Island property tax credit eligibility status"
    definition_period = YEAR
    reference = (
        "http://webserver.rilin.state.ri.us/Statutes/TITLE44/44-33/44-33-3.htm"
    )
    defined_for = StateCode.RI

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ri.tax.income.credits.property_tax
        # minimum age eligibility
        head_age = tax_unit("age_head", period)
        spouse_age = tax_unit("age_spouse", period)
        age_eligible = max_(head_age, spouse_age) >= p.age_threshold
        # disability eligibility
        head_is_disabled = tax_unit("head_is_disabled", period)
        spouse_is_disabled = tax_unit("spouse_is_disabled", period)
        is_disabled = head_is_disabled | spouse_is_disabled
        return age_eligible | is_disabled



from fiscalsim_us.model_api import *


class is_ssi_eligible_individual(Variable):
    value_type = bool
    entity = Person
    label = "Is an SSI-eligible individual"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/42/1382#a"

    def formula(person, period, parameters):
        aged_blind_disabled = person("is_ssi_aged_blind_disabled", period)
        is_ssi_eligible_spouse = person("is_ssi_eligible_spouse", period)
        return aged_blind_disabled & ~is_ssi_eligible_spouse



from fiscalsim_us.model_api import *


class is_person_demographic_tanf_eligible(Variable):
    value_type = bool
    entity = Person
    definition_period = YEAR
    label = "Person-level eligiblity for TANF based on age, pregnancy, etc."

    def formula(person, period, parameters):
        child_0_17 = person("is_child", period)
        is_18 = person("age", period) == 18
        full_time_student = person("is_full_time_student", period)
        school_enrolled_18_year_old = full_time_student & is_18
        pregnant = person("is_pregnant", period)
        return child_0_17 | school_enrolled_18_year_old | pregnant



from .cps import (
    CPS_2020,
    CPS_2021,
    CPS_2022,
    CPS_2023,
    RawCPS_2020,
    RawCPS_2021,
    RawCPS_2022,
    EnhancedCPS_2023,
    PUFExtendedCPS_2023,
    CalibratedCPS_2023,
)

from .poverty_tracker.poverty_tracker import PovertyTracker

DATASETS = [
    CPS_2020,
    CPS_2021,
    CPS_2022,
    CPS_2023,
    EnhancedCPS_2023,
    PUFExtendedCPS_2023,
    CalibratedCPS_2023,
    PovertyTracker,
]



import torch
import pandas as pd
import numpy as np
from fiscalsim_us import Microsimulation
import plotly.express as px
from tqdm import tqdm
from policyengine_core.data import Dataset
from fiscalsim_us.data import CPS_2023
import numpy as np
from fiscalsim_us import Microsimulation
import pandas as pd

person_df = pd.read_csv("puf_imputed_cps_person.csv.gz", compression="gzip")

FINANCIAL_VARIABLES = [
    "employment_income",
    "self_employment_income",
    "partnership_s_corp_income",
    "farm_income",
    "farm_rent_income",
    "short_term_capital_gains",
    "long_term_capital_gains",
    "taxable_interest_income",
    "tax_exempt_interest_income",
    "rental_income",
    "qualified_dividend_income",
    "non_qualified_dividend_income",
    "taxable_pension_income",
    "social_security",
]


class PUFExtendedCPS(Dataset):
    name = "puf_extended_cps"
    label = "PUF-extended CPS"
    file_path = "puf_extended_cps.h5"
    data_format = Dataset.ARRAYS
    time_period = "2023"

    def generate(self):
        new_data = {}
        cps = CPS_2023()
        cps_data = cps.load()
        for variable in list(set(cps.variables) | set(FINANCIAL_VARIABLES)):
            if "_id" in variable:
                # Append on a copy multiplied by 10
                new_data[variable] = np.concatenate(
                    [cps_data[variable][...], cps_data[variable][...] + 1e8]
                )
            elif "_weight" in variable:
                # Append on a zero-weighted copy
                new_data[variable] = np.concatenate(
                    [
                        cps_data[variable][...],
                        np.zeros_like(cps_data[variable][...]),
                    ]
                )
            else:
                # Append on a copy
                if variable in FINANCIAL_VARIABLES:
                    if variable not in cps.variables:
                        if variable == "social_security":
                            ## SS is an edge case
                            original_values = (
                                cps_data["social_security_retirement"][...]
                                + cps_data["social_security_disability"][...]
                            )
                        else:
                            original_values = np.zeros_like(
                                cps_data["employment_income"][...]
                            )
                    else:
                        original_values = cps_data[variable][...]
                    new_data[variable] = np.concatenate(
                        [original_values, person_df[variable].values]
                    )
                else:
                    new_data[variable] = np.concatenate(
                        [cps_data[variable][...], cps_data[variable][...]]
                    )

        self.save_dataset(new_data)


puf_extended_cps = PUFExtendedCPS()
puf_extended_cps.generate()


simulation = Microsimulation(dataset=puf_extended_cps)
TIME_PERIOD = 2023
simulation.default_calculation_period = TIME_PERIOD
parameters = simulation.tax_benefit_system.parameters.calibration(
    f"{TIME_PERIOD}-01-01"
)

household_weights = torch.tensor(
    simulation.calculate("household_weight").values, dtype=torch.float32
)
weight_adjustment = torch.tensor(
    np.random.random(household_weights.shape) * 10,
    requires_grad=True,
    dtype=torch.float32,
)

values_df = pd.DataFrame()
targets = {}
equivalisation = {}

# We need to normalise the targets. Common regression targets are often 1e1 to 1e3 (this informs the scale of the learning rate).
COUNT_HOUSEHOLDS = household_weights.sum().item()
FINANCIAL_EQUIVALISATION = COUNT_HOUSEHOLDS
POPULATION_EQUIVALISATION = COUNT_HOUSEHOLDS / 1e5

# Financial totals
"""
AGI_VARIABLES = [
    "adjusted_gross_income",
    "employment_income",
    # "taxable_interest_and_ordinary_dividends", Doesn't exist yet
    "qualified_dividend_income",
    "net_capital_gain",
    "self_employment_income",
    "taxable_pension_income",
    "taxable_social_security",
    # "irs_other_income", Doesn't exist yet
    # "above_the_line_deductions",
]


for variable_name in AGI_VARIABLES:
    label = simulation.tax_benefit_system.variables[variable_name].label + " aggregate"
    values_df[label] = simulation.calculate(
        variable_name, map_to="household"
    ).values
    targets[label] = parameters.gov.cbo.income_by_source[variable_name]
    equivalisation[label] = FINANCIAL_EQUIVALISATION
"""

# Above is commented because we'll use IRS SOI aggregates uprated by CBO forecasts, rather than just CBO forecasts (SOI aggregates have more detail)

for variable_name in FINANCIAL_VARIABLES:
    label = (
        simulation.tax_benefit_system.variables[variable_name].label
        + " aggregate"
    )
    values_df[label] = simulation.calculate(
        variable_name, map_to="household"
    ).values
    targets[label] = parameters.gov.irs.soi[variable_name]
    equivalisation[label] = FINANCIAL_EQUIVALISATION


# Program spending from CBO baseline projections

PROGRAMS = [
    "income_tax",
    "snap",
    "social_security",
    "ssi",
    "unemployment_compensation",
]

for variable_name in PROGRAMS:
    label = simulation.tax_benefit_system.variables[variable_name].label
    values_df[label] = simulation.calculate(
        variable_name, map_to="household"
    ).values
    targets[label] = parameters.gov.cbo[variable_name]
    equivalisation[label] = FINANCIAL_EQUIVALISATION

snap_participation = parameters.gov.usda.snap.participation
ssi_participation = parameters.gov.ssa.ssi.participation
ss_participation = parameters.gov.ssa.social_security.participation

for program, participation in zip(
    ["snap", "ssi", "social_security"],
    [snap_participation, ssi_participation, ss_participation],
):
    label = simulation.tax_benefit_system.variables[program].label
    entity_level = simulation.tax_benefit_system.variables[program].entity.key
    entity_level_value = simulation.calculate(program)
    values_df[f"{label} participants"] = simulation.map_result(
        entity_level_value > 0, entity_level, "household"
    )
    targets[f"{label} participants"] = participation
    equivalisation[f"{label} participants"] = POPULATION_EQUIVALISATION


# Number of tax returns by AGI size

agi_returns_thresholds = (
    parameters.gov.irs.soi.agi.number_of_returns.thresholds
)
agi_returns_values = parameters.gov.irs.soi.agi.number_of_returns.amounts
agi = simulation.calculate("adjusted_gross_income").values
is_filer = simulation.calculate("income_tax").values != 0
for i in range(len(agi_returns_thresholds)):
    lower = agi_returns_thresholds[i]
    if i == len(agi_returns_thresholds) - 1:
        upper = np.inf
    else:
        upper = agi_returns_thresholds[i + 1]

    in_range = (agi >= lower) * (agi < upper) * is_filer
    household_returns_in_range = simulation.map_result(
        in_range, "tax_unit", "household"
    )

    name = f"Tax returns with ${lower:,.0f} <= AGI < ${upper:,.0f}"
    values_df[name] = household_returns_in_range
    targets[name] = agi_returns_values[i]
    equivalisation[name] = POPULATION_EQUIVALISATION
# Total AGI aggregate by AGI band

agi_returns_thresholds = parameters.gov.irs.soi.agi.total_agi.thresholds
agi_returns_values = parameters.gov.irs.soi.agi.total_agi.amounts
for i in range(len(agi_returns_thresholds)):
    lower = agi_returns_thresholds[i]
    if i == len(agi_returns_thresholds) - 1:
        upper = np.inf
    else:
        upper = agi_returns_thresholds[i + 1]

    in_range = (agi >= lower) * (agi < upper) * is_filer
    agi_in_range = agi * in_range
    household_agi_in_range = simulation.map_result(
        agi_in_range, "tax_unit", "household"
    )

    name = (
        f"Total AGI from tax returns with ${lower:,.0f} <= AGI < ${upper:,.0f}"
    )
    values_df[name] = household_agi_in_range
    targets[name] = agi_returns_values[i]
    equivalisation[name] = FINANCIAL_EQUIVALISATION

# Total population
values_df["U.S. population"] = simulation.calculate(
    "people", map_to="household"
).values
targets["U.S. population"] = parameters.populations.total
equivalisation["U.S. population"] = POPULATION_EQUIVALISATION

# Population by 10-year age group and sex
age = simulation.calculate("age").values
is_male = simulation.calculate("is_male")
for lower_age_group in range(0, 90, 10):
    for possible_is_male in (True, False):
        in_age_range = (age >= lower_age_group) & (age < lower_age_group + 5)
        in_sex_category = is_male == possible_is_male
        count_people_in_range = simulation.map_result(
            in_age_range * in_sex_category, "person", "household"
        )
        sex_category = "male" if possible_is_male else "female"
        name = f"{lower_age_group} to {lower_age_group + 5} and {sex_category} population"
        values_df[name] = count_people_in_range
        targets[name] = (
            household_weights.numpy() * count_people_in_range
        ).sum()
        equivalisation[name] = POPULATION_EQUIVALISATION

# Household population by number of adults and children

household_count_adults = simulation.map_result(
    age >= 18, "person", "household"
)
household_count_children = simulation.map_result(
    age < 18, "person", "household"
)

for count_adults in range(1, 3):
    for count_children in range(0, 4):
        in_criteria = (
            (household_count_adults == count_adults)
            * (household_count_children == count_children)
            * 1.0
        )
        name = f"{count_adults}-adult, {count_children}-child household population"
        values_df[name] = in_criteria
        targets[name] = (household_weights.numpy() * in_criteria).sum()
        equivalisation[name] = POPULATION_EQUIVALISATION

# Tax filing unit counts by filing status

filing_status = simulation.calculate("filing_status").values

for filing_status_value in np.unique(filing_status):
    is_filing_status = filing_status == filing_status_value
    name = f"Filing status {filing_status_value.lower()} population"
    household_filing_status_unit_counts = simulation.map_result(
        is_filing_status, "tax_unit", "household"
    )
    values_df[name] = household_filing_status_unit_counts
    targets[name] = (
        household_weights.numpy() * household_filing_status_unit_counts
    ).sum()
    equivalisation[name] = POPULATION_EQUIVALISATION

targets_array = torch.tensor(list(targets.values()), dtype=torch.float32)
equivalisation_factors_array = torch.tensor(
    list(equivalisation.values()), dtype=torch.float32
)


def aggregate(
    adjusted_weights: torch.Tensor, values: pd.DataFrame
) -> torch.Tensor:
    broadcasted_weights = adjusted_weights.reshape(-1, 1)
    weighted_values = torch.matmul(
        broadcasted_weights.T, torch.tensor(values.values, dtype=torch.float32)
    )
    return weighted_values


training_log_df = pd.DataFrame()

progress_bar = tqdm(range(250_000), desc="Calibrating weights")
for i in progress_bar:
    adjusted_weights = torch.relu(household_weights + weight_adjustment)
    result = (
        aggregate(adjusted_weights, values_df) / equivalisation_factors_array
    )
    loss = torch.mean(
        (result - targets_array / equivalisation_factors_array) ** 2
    )
    loss.backward()
    if i % 20 == 0:
        current_loss = loss.item()
        progress_bar.set_description_str(
            f"Calibrating weights | Loss = {current_loss:,.0f}"
        )
        current_aggregates = (
            (result * equivalisation_factors_array).detach().numpy()[0]
        )
        training_log_df = pd.concat(
            [
                training_log_df,
                pd.DataFrame(
                    {
                        "name": list(targets.keys()) + ["total"],
                        "epoch": [i] * len(targets) + [i],
                        "value": list(current_aggregates) + [current_loss],
                        "target": list(targets.values()) + [0],
                    }
                ),
            ]
        )
    weight_adjustment.data -= 1e-1 * weight_adjustment.grad
    weight_adjustment.grad.zero_()

training_log_df.to_csv("training_log.csv.gz", compression="gzip")

from fiscalsim_us.data.storage import STORAGE_FOLDER


class EnhancedCPS(Dataset):
    name = "enhanced_cps"
    label = "Enhanced CPS"
    file_path = STORAGE_FOLDER / "enhanced_cps.h5"
    data_format = Dataset.ARRAYS
    time_period = "2023"

    def generate(self):
        new_data = {}
        cps = PUFExtendedCPS()
        cps_data = cps.load()
        for variable in cps.variables:
            if variable == "household_weight":
                new_data[variable] = adjusted_weights.detach().numpy()
            elif "_weight" not in variable:
                new_data[variable] = cps_data[variable][...]

        self.save_dataset(new_data)


enhanced_cps = EnhancedCPS()
enhanced_cps.generate()



from fiscalsim_us.model_api import *


class mt_nonrefundable_credits(Variable):
    """
    Line 19 of Montana state individual tax return form 2
    """

    value_type = float
    entity = TaxUnit
    label = "Montana refundable credits"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT

    adds = "gov.states.mt.tax.income.credits.non_refundable"



from fiscalsim_us.model_api import *


class dc_disabled_exclusion_subtraction(Variable):
    value_type = float
    entity = Person
    label = "DC disabled exclusion subtraction"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://otr.cfo.dc.gov/sites/default/files/dc/sites/otr/publication/attachments/52926_D-40_12.21.21_Final_Rev011122.pdf#page=63"
        "https://otr.cfo.dc.gov/sites/default/files/dc/sites/otr/publication/attachments/2021_D-2440.pdf"
        "https://otr.cfo.dc.gov/sites/default/files/dc/sites/otr/publication/attachments/2022_D-40_Booklet_Final_blk_01_23_23_Ordc.pdf#page=55"
        "https://otr.cfo.dc.gov/sites/default/files/dc/sites/otr/publication/attachments/2022_D-2440.pdf"
    )
    defined_for = StateCode.DC

    def formula(person, period, parameters):
        # determine disablity-related eligibility
        is_disabled = person("is_permanently_and_totally_disabled", period)
        gets_ssi = person("ssi", period) > 0
        gets_ssdi = person("social_security_disability", period) > 0
        disabled_eligible = is_disabled & (gets_ssi | gets_ssdi)
        # determine income-related eligibility
        INCOME_SOURCES = ["household_market_income", "household_benefits"]
        household_income = add(person.household, period, INCOME_SOURCES)
        p = parameters(period).gov.states.dc.tax.income.subtractions
        income_eligible = household_income < p.disabled_exclusion.income_limit
        # return subtraction amount if meet both eligibilty requirements
        subtraction_amount = p.disabled_exclusion.amount
        return (disabled_eligible & income_eligible) * subtraction_amount



from fiscalsim_us.model_api import *


class wa_capital_gains_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "Washington capital gains tax"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.WA

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.wa.tax.income.capital_gains
        ltcg = add(tax_unit, period, ["long_term_capital_gains"])
        # Deduct charitable contributions.
        charitable_contributions = add(
            tax_unit,
            period,
            ["charitable_cash_donations", "charitable_non_cash_donations"],
        )
        charitable_contributions_above_exemption = max_(
            0, charitable_contributions - p.deductions.charitable.exemption
        )
        charitable_deduction = min_(
            charitable_contributions_above_exemption,
            p.deductions.charitable.cap,
        )
        # Get base after standard and charitable deductions.
        taxable_ltcg = max_(
            0, ltcg - charitable_deduction - p.deductions.standard
        )
        return taxable_ltcg * p.rate



from fiscalsim_us.model_api import *


class mt_total_payments(Variable):
    """
    Line 25 on Montana individual income tax retrun form 2
    """

    value_type = float
    entity = TaxUnit
    label = "Montana total payments"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT

    adds = ["mt_refundable_credits", "mt_eitc"]
    subtracts = ["mt_contributions_penalties_interest"]



from fiscalsim_us.model_api import *


class pell_grant_head_available_income(Variable):
    value_type = float
    entity = Person
    label = "Pell Grant head available income"
    definition_period = YEAR

    def formula(person, period, parameters):
        assets = person.tax_unit("pell_grant_head_assets", period)
        simplified = person("pell_grant_simplified_formula_applies", period)
        allowances = person("pell_grant_head_allowances", period)
        income = person.tax_unit("pell_grant_primary_income", period)
        formula = person("pell_grant_formula", period)
        p = parameters(period).gov.ed.pell_grant.efc.head
        adjusted_assets = (
            assets * p.asset_assessment_rate[formula] * ~simplified
        )
        adjusted_income = (income - allowances) * p.income_assessment_rate[
            formula
        ]
        return adjusted_income + adjusted_assets
